import typing as _tp
from ._typing import Indexable as _Indexable
from ._jacobi import jacobi
import numpy as np


def propagate(
    fn: _tp.Callable,
    x: _tp.Union[float, _Indexable[float]],
    cov: _tp.Union[float, _Indexable[float], _Indexable[_Indexable[float]]],
    *args,
    **kwargs,
) -> _tp.Tuple[np.ndarray, np.ndarray]:
    """
    Numerically propagates the covariance of function inputs to function outputs.

    The function computes C' = J C J^T, where C is the covariance matrix of the input,
    C' the matrix of the output, and J is the Jacobi matrix of first derivatives of the
    mapping function fn. The Jacobi matrix is computed numerically.

    Parameters
    ----------
    fn: callable
        Function that computes r = fn(x, [y, ...]). The arguments of the function are
        each allowed to be scalars or one-dimensional arrays. If the function accepts
        several arguments, their uncertainties are treated as uncorrelated.
        Functions that accept several correlated arguments must be wrapped, see examples.
        The result of the function may be a scalar or a one-dimensional array with a
        different lenth as the input.
    x: float or array-like with shape (N,)
        Input vector. An array-like is converted before passing it to the callable.
    cov: float or array-like with shape (N,) or shape(N, N)
        Covariance matrix of input vector. If the array is one-dimensional, it is
        interpreted as the diagonal of a covariance matrix with zero off-diagonal
        elements.
    *args:
        If the function accepts several arguments that are mutually independent, these
        is possible to pass those values and covariance matrices pairwise, see examples.
    **kwargs:
        Extra arguments are passed to :func:`jacobi`.

    Returns
    -------
    y, ycov
        y is the result of fn(x).
        ycov is the propagated covariance matrix.
        If ycov is a matrix, unless y is a number. In that case, ycov is also
        reduced to a number.

    Examples
    --------
    General error propagation maps input vectors to output vectors::

        def fn(x):
            return x ** 2 + 1

        x = [1, 2]
        xcov = [[3, 1],
                [1, 4]]

        y, ycov = propagate(fn, x, xcov)

    In the previous example, the function y = fn(x) treats all x values independently,
    so the Jacobian computed from fn(x) has zero off-diagonal entries. In this case,
    one can speed up the calculation significantly with a special keyword::

        # same result as before, but faster and uses much less memory
        y, ycov = propagate(fn, x, xcov, diagonal=True)

    If the function accepts several arguments, their uncertainties are treated as
    uncorrelated::

        def fn(x, y):
            return x + y

        x = 1
        y = 2
        xcov = 2
        ycov = 3

        z, zcov = propagate(fn, x, xcov, y, ycov)

    Functions that accept several correlated arguments must be wrapped::

        def fn(x, y):
            return x + y

        x = 1
        y = 2
        sigma_x = 3
        sigma_y = 4
        rho_xy = 0.5

        r = [x, y]
        cov_xy = rho_xy * sigma_x * sigma_y
        rcov = [[sigma_x ** 2, cov_xy], [cov_xy, sigma_y ** 2]]

        def fn_wrapped(r):
            return fn(r[0], r[1])

        z, zcov = propagate(fn_wrapped, r, rcov)

    See Also
    --------
    jacobi
    """
    if args:
        if len(args) % 2 != 0:
            raise ValueError("number of extra positional arguments must be even")

        args_a = [np.asarray(_) for _ in ((x, cov) + args)]
        x_parts = args_a[::2]
        cov_parts = args_a[1::2]
        y_a = np.asarray(fn(*x_parts))
        return _propagate_independent(fn, y_a, x_parts, cov_parts, **kwargs)

    x_a = np.asarray(x)
    cov_a = np.asarray(cov)
    y_a = np.asarray(fn(x_a))

    if x_a.ndim > 1:
        raise ValueError("x must have dimension 0 or 1")

    if kwargs.get("diagonal", False):
        return _propagate_diagonal(fn, y_a, x_a, cov_a, **kwargs)

    if cov_a.ndim > 2:
        raise ValueError("cov must have dimension 0, 1, or 2")

    return _propagate_full(fn, y_a, x_a, cov_a, **kwargs)


def _propagate_full(fn, y: np.ndarray, x: np.ndarray, xcov: np.ndarray, **kwargs):
    x_a = np.atleast_1d(x)

    _check_x_xcov_compatibility(x_a, xcov)

    jac = np.asarray(jacobi(fn, x_a, **kwargs)[0])

    y_len = len(y) if y.ndim == 1 else 1

    if jac.ndim == 1:
        jac = jac.reshape((y_len, len(x_a)))
    assert np.ndim(jac) == 2

    ycov = _jac_cov_product(jac, xcov)

    if y.ndim == 0:
        ycov = np.squeeze(ycov)

    return y, ycov


def _propagate_diagonal(fn, y: np.ndarray, x: np.ndarray, xcov: np.ndarray, **kwargs):
    x_a = np.atleast_1d(x)

    _check_x_xcov_compatibility(x_a, xcov)

    jac = np.asarray(jacobi(fn, x_a, **kwargs)[0])
    assert jac.ndim <= 1

    ycov = _jac_cov_product(jac, xcov)

    if y.ndim == 0:
        assert ycov.ndim == 0

    return y, ycov


def _propagate_independent(
    fn,
    y: np.ndarray,
    x_parts: _tp.List[np.ndarray],
    xcov_parts: _tp.List[np.ndarray],
    **kwargs,
):
    ycov = 0

    for i, x in enumerate(x_parts):
        rest = x_parts[:i] + x_parts[i + 1 :]

        def wrapped(x, *rest):
            args = rest[:i] + (x,) + rest[i:]
            return fn(*args)

        x_a = np.atleast_1d(x)
        xcov = xcov_parts[i]
        _check_x_xcov_compatibility(x_a, xcov)

        jac = np.asarray(jacobi(wrapped, x_a, *rest, **kwargs)[0])
        ycov += _jac_cov_product(jac, xcov)

    if y.ndim == 0:
        ycov = np.squeeze(ycov)

    return y, ycov


def _jac_cov_product(jac: np.ndarray, xcov: np.ndarray):
    if xcov.ndim == 2:
        return np.einsum(
            "i,j,ij -> ij" if jac.ndim == 1 else "ij,kl,jl", jac, jac, xcov
        )
    elif jac.ndim == 2:
        if xcov.ndim == 1:
            return np.einsum("ij,kj,j", jac, jac, xcov)
        return np.einsum("ij,kj", jac, jac) * xcov
    assert jac.ndim < 2 and xcov.ndim < 2
    return xcov * jac**2


def _check_x_xcov_compatibility(x: np.ndarray, xcov: np.ndarray):
    if xcov.ndim > 0 and len(xcov) != (len(x) if x.ndim == 1 else 1):
        # this works for 1D and 2D xcov
        raise ValueError("x and cov have incompatible shapes")



from jacobi import jacobi, propagate
import numpy as np
import pytest


@pytest.mark.parametrize("diagonal", (False, True))
def test_jacobi(diagonal, benchmark):
    x = np.linspace(0, 1, 100)

    def fn(x):
        return x**2 + 3 * x + 1

    def run():
        jacobi(fn, x, diagonal=diagonal)

    benchmark(run)


@pytest.mark.parametrize("diagonal", (False, True))
def test_propagate(diagonal, benchmark):
    x = np.linspace(0, 1, 100)

    def fn(x):
        return x**2 + 3 * x + 1

    def run():
        propagate(fn, x, x, diagonal=diagonal)

    benchmark(run)


@pytest.mark.parametrize("diagonal", (False, True))
def test_propagate_2dcov(diagonal, benchmark):
    x = np.linspace(0, 1, 100)

    def fn(x):
        return x**2 + 3 * x + 1

    def run():
        propagate(fn, x, np.diag(x), diagonal=diagonal)

    benchmark(run)



#!/usr/bin/env python3
import subprocess as subp
from pathlib import Path
import sys

cmd = "sphinx-build -W -b html -d _build/doctrees . _build/html"

sys.exit(subp.call(cmd.split(), cwd=Path(__file__).parent))



import numpy as np
import typing as _tp
from ._typing import Indexable as _Indexable


def jacobi(
    fn: _tp.Callable,
    x: _tp.Union[float, _Indexable[float]],
    *args,
    diagonal: bool = False,
    method: int = None,
    mask: np.ndarray = None,
    rtol: float = 0,
    maxiter: int = 10,
    maxgrad: int = 3,
    step: _tp.Tuple[float, float] = None,
    diagnostic: dict = None,
) -> _tp.Tuple[np.ndarray, np.ndarray]:
    """
    Return first derivative and its error estimate.

    Parameters
    ----------
    fn : Callable
        Function with the signature `fn(x, *args)`, where `x` is a number or an
        array of numbers and `*args` are optional auxiliary arguments.
    x : Number or array of numbers
        The derivative is computed with respect to `x`. If `x` is an array, the Jacobi
        matrix is computed with respect to each element of `x`.
    *args : tuple
        Additional arguments passed to the function.
    diagonal : boolean, optional
        If it is known that the Jacobian computed from the function contains has
        off-diagonal entries that are all zero, the calculation can be speed up
        significantly. Set this to true to only compute the diagonal entries of the
        Jacobi matrix, which are returned as a 1D array. This is faster and uses much
        less memory if the vector x is very large. Default is False.
    method : {-1, 0, 1} or None, optional
        Whether to compute central (0), forward (1) or backward derivatives (-1).
        The default (None) uses auto-detection.
    mask : array or None, optional
        If `x` is an array and `mask` is not None, compute the Jacobi matrix only for the
        part of the array selected by the mask.
    rtol : float, optional
        Relative tolerance for the derivative. The algorithm stops when this relative
        tolerance is reached. If 0 (the default), the algorithm iterates until the
        error estimate of the derivative does not improve further.
    maxiter : int, optional
        Maximum number of iterations of the algorithm.
    maxgrad : int, optional
        Maximum grad of the extrapolation polynomial.
    step : tuple of float or None, optional
        Factors that reduce the step size in each iteration relative to the previous
        step.
    diagnostic : dict or None, optional
        If an empty dict is passed to this keyword, it is filled with diagnostic
        information produced by the algorithm. This reduces performance and is only
        intended for debugging.

    Returns
    -------
    array, array
        Derivative and its error estimate.
    """
    if diagonal:
        # TODO maybe solve this without introducing a wrapper function
        return jacobi(
            lambda dx, x: fn(x + dx),
            0,
            x,
            method=method,
            rtol=rtol,
            maxiter=maxiter,
            maxgrad=maxgrad,
            step=step,
            diagnostic=diagnostic,
        )

    if maxiter <= 0:
        raise ValueError("maxiter must be > 0")
    if maxgrad < 0:
        raise ValueError("maxgrad must be >= 0")
    if step is not None:
        if not (0 < step[0] < 0.5):
            raise ValueError("step[0] must be between 0 and 0.5")
        if not (0 < step[1] < 1):
            raise ValueError("step[1] must be between 0 and 1")

    squeeze = np.ndim(x) == 0
    x = np.atleast_1d(x).astype(float)
    assert x.ndim == 1

    x_indices = np.arange(len(x))
    if mask is not None:
        x_indices = x_indices[mask]
    nx = len(x_indices)

    if diagnostic is not None:
        diagnostic["method"] = np.zeros(nx, dtype=np.int8)
        diagnostic["iteration"] = np.zeros(len(x_indices), dtype=np.uint8)

    if method is not None and method not in (-1, 0, 1):
        raise ValueError("method must be -1, 0, 1")

    f0 = None
    jac = None
    err = None
    for ik, k in enumerate(x_indices):
        # if step is None, use optimal step sizes for central derivatives
        h = _steps(x[k], step or (0.25, 0.5), maxiter)
        # if method is None, auto-detect for each x[k]
        md, f0, r = _first(method, f0, fn, x, k, h[0], args)

        if diagnostic is not None:
            diagnostic["method"][ik] = md

        if md != 0 and step is None:
            # optimal step sizes for forward derivatives
            h = _steps(x[k], (0.125, 0.125), maxiter)

        r_shape = np.shape(r)
        r = np.reshape(r, -1)
        nr = len(r)
        re = np.full(nr, np.inf)
        todo = np.ones(nr, dtype=bool)
        fd = [r]

        if jac is None:
            jac = np.empty(r_shape + (nx,), dtype=r.dtype)
            err = np.empty(r_shape + (nx,), dtype=r.dtype)
            if diagnostic is not None:
                diagnostic["call"] = np.zeros((nr, nx), dtype=np.uint8)

        if diagnostic is not None:
            diagnostic["call"][:, ik] = 2 if md == 0 else 3

        for i in range(1, len(h)):
            fdi = _derive(md, f0, fn, x, k, h[i], args)
            fdi = np.reshape(fdi, -1)
            fd.append(fdi if i == 1 else fdi[todo])
            if diagnostic is not None:
                diagnostic["call"][todo, ik] += 2
                diagnostic["iteration"][ik] += 1

            # polynomial fit with one extra degree of freedom;
            # use latest maxgrad + 1 data points
            grad = min(i - 1, maxgrad)
            start = i - (grad + 1)
            stop = i + 1
            q, c = np.polyfit(
                h[start:stop] ** 2, fd[start:], grad, rcond=None, cov=True
            )
            ri = q[-1]
            # pulls have roughly unit variance, however,
            # the pull distribution is not gaussian and looks
            # more like student's t
            rei = c[-1, -1] ** 0.5

            # update estimates that have significantly smaller error
            sub_todo = rei < re[todo]
            todo1 = todo.copy()
            todo[todo1] = sub_todo
            r[todo] = ri[sub_todo]
            re[todo] = rei[sub_todo]

            # do not improve estimates further which meet the tolerance
            if rtol > 0:
                sub_todo &= rei > rtol * np.abs(ri)
                todo[todo1] = sub_todo

            if np.sum(todo) == 0:
                break

            # shrink previous vectors of estimates
            fd = [fdi[sub_todo] for fdi in fd]

        jac[..., ik] = r.reshape(r_shape)
        err[..., ik] = re.reshape(r_shape)

    if diagnostic is not None:
        diagnostic["call"].shape = r_shape + (nx,)

    if squeeze:
        if diagnostic is not None:
            diagnostic["call"] = np.squeeze(diagnostic["call"])
        jac = np.squeeze(jac)
        err = np.squeeze(err)

    return jac, err


def _steps(p, step, maxiter):
    h0, factor = step
    h = p * h0
    if h == 0:
        h = h0
    return h * factor ** np.arange(maxiter)


def _derive(mode, f0, f, x, i, h, args):
    x1 = x.copy()
    x2 = x.copy()
    if mode == 0:
        x1[i] += h
        x2[i] -= h
        return (f(x1, *args) - f(x2, *args)) * (0.5 / h)
    h = h * mode
    x1[i] += h
    x2[i] += 2 * h
    f1 = f(x1, *args)
    f2 = f(x2, *args)
    return (-3 * f0 + 4 * f1 - f2) * (0.5 / h)


def _first(method, f0, f, x, i, h, args):
    norm = 0.5 / h
    f1 = None
    f2 = None
    if method is None or method == 0:
        x1 = x.copy()
        x2 = x.copy()
        x1[i] -= h
        x2[i] += h
        f1 = f(x1, *args)
        f2 = f(x2, *args)
        if method is None:
            if np.any(np.isnan(f1)):  # forward method
                method = 1
            elif np.any(np.isnan(f2)):  # backward method
                method = -1
            else:
                method = 0
    if method == 0:
        return method, None, (f1 - f2) * norm
    if f0 is None:
        f0 = f(x, *args)
    if method == -1:
        h = -h
        norm = -norm
    if f1 is None:
        x1 = x.copy()
        x1[i] += h
        f1 = f(x1, *args)
    elif method == 1:
        f1 = f2
    x2 = x.copy()
    x2[i] += 2 * h
    f2 = f(x2, *args)
    return method, f0, (-3 * f0 + 4 * f1 - f2) * norm



import numpy as np
from numpy.testing import assert_allclose
from jacobi import propagate
import pytest


def test_00():
    def fn(x):
        return x**2 + 5

    x = 2
    xcov = 3
    y, ycov = propagate(fn, x, xcov)
    assert_allclose(y, fn(x))
    assert_allclose(ycov, (2 * x) ** 2 * xcov)


def test_01():
    def fn(x):
        return np.ones(2) * x**2

    x = 2
    xcov = 3
    y, ycov = propagate(fn, x, xcov)
    assert_allclose(y, fn(x))
    jac = (2 * x) * np.ones((2, 1))
    xcov2 = np.atleast_2d(xcov)
    assert_allclose(ycov, np.linalg.multi_dot([jac, xcov2, jac.T]))

    with pytest.raises(ValueError):
        propagate(fn, x, np.ones(2))

    with pytest.raises(ValueError):
        propagate(fn, x, np.ones((2, 2)))


def test_10():
    def fn(x):
        return np.sum(x)

    x = np.arange(2)
    for xcov in ([1, 1], [[1, 0], [0, 1]]):

        y, ycov = propagate(fn, x, xcov)
        assert_allclose(y, fn(x))
        jac = np.ones((1, len(x)))

        if np.ndim(xcov) == 1:
            xcov2 = np.zeros((2, 2))
            for i in range(2):
                xcov2[i, i] = xcov[i]
        else:
            xcov2 = xcov
        assert_allclose(ycov, np.linalg.multi_dot([jac, xcov2, jac.T]))
        assert np.ndim(y) == 0
        assert np.ndim(ycov) == 0


def test_11():
    A = np.array([(1, 2), (3, 4), (5, 6)], dtype=float)

    def fn(x):
        return np.dot(A, x)

    x = np.arange(2)

    for xcov in ([1, 1], [[1, 0], [0, 1]]):
        y, ycov = propagate(fn, x, xcov)
        assert_allclose(y, fn(x))
        jac = A
        if np.ndim(xcov) == 1:
            xcov2 = np.zeros((2, 2))
            for i in range(2):
                xcov2[i, i] = xcov[i]
        else:
            xcov2 = xcov
        assert_allclose(ycov, np.linalg.multi_dot([jac, xcov2, jac.T]))


@pytest.mark.parametrize("ndim", (1, 2))
def test_cov_1d_2d(ndim):
    def fn(x):
        return x

    x = [1, 2]
    xcov_1d = [3, 4]
    xcov_2d = np.diag(xcov_1d)

    y, ycov = propagate(fn, x, xcov_1d if ndim == 1 else xcov_2d)

    assert np.ndim(ycov) == 2

    assert_allclose(y, x)
    assert_allclose(ycov, xcov_2d)


def test_two_arguments_1():
    def fn1(x, y):
        return (x - y) / (x + y)

    x = 1
    xcov = 2
    y = 3
    ycov = 4

    z, zcov = propagate(fn1, x, xcov, y, ycov)
    assert z.ndim == 0
    assert zcov.ndim == 0

    def fn2(r):
        return fn1(r[0], r[1])

    r = [x, y]
    rcov = np.diag([xcov, ycov])

    z_ref, zcov_ref = propagate(fn2, r, rcov)

    assert_allclose(z_ref, z)
    assert_allclose(zcov_ref, zcov)


def test_two_arguments_2():
    def fn1(x, y):
        return np.concatenate([x, np.atleast_1d(y)])

    x = [1, 2]
    xcov = [2, 3]
    y = 3
    ycov = 4

    z, zcov = propagate(fn1, x, xcov, y, ycov)

    def fn2(r):
        return fn1(r[:2], r[2])

    r = [*x, y]
    rcov = np.diag([*xcov, ycov])

    z_ref, zcov_ref = propagate(fn2, r, rcov)

    assert_allclose(z, z_ref)
    assert_allclose(zcov, zcov_ref)


def test_bad_number_of_arguments():
    with pytest.raises(ValueError, match="number of extra"):
        propagate(lambda x: x, 1, 2, 3)


def test_bad_input_dimensions():
    def fn(x):
        return x

    with pytest.raises(ValueError):
        propagate(fn, [[1]], 1)

    with pytest.raises(ValueError):
        propagate(fn, 1, [[[1]]])


def test_diagonal_1():
    def fn(x):
        return x**2 + 3

    def fprime(x):
        x = np.atleast_1d(x)
        return 2 * x

    x = 2
    xcov = 3

    y, ycov = propagate(fn, x, xcov, diagonal=True)

    assert ycov.ndim == 0
    assert_allclose(ycov, fprime(x) ** 2 * xcov)


def test_diagonal_2():
    def fn(x):
        return x**2 + 3

    def fprime(x):
        x = np.atleast_1d(x)
        return 2 * x

    x = [1, 2]
    xcov = [3, 4]

    y, ycov = propagate(fn, x, xcov, diagonal=True)

    x_a = np.atleast_1d(x)
    assert ycov.ndim == 1
    assert_allclose(y, fn(x_a))

    jac = np.diag(fprime(x_a))
    cov_a = np.diag(xcov)
    ycov_ref = np.linalg.multi_dot((jac, cov_a, jac.T))
    assert ycov_ref[0, 1] == 0
    assert ycov_ref[1, 0] == 0
    assert_allclose(ycov, np.diag(ycov_ref))


def test_diagonal_3():
    def fn(x):
        return x**2 + 1

    x = [1, 2]
    xcov = [[3, 1], [1, 4]]

    y, ycov = propagate(fn, x, xcov, diagonal=True)

    y_ref, ycov_ref = propagate(fn, x, xcov)

    assert_allclose(y, y_ref)
    assert_allclose(ycov, ycov_ref)



from typing import Iterable, Sized, Generic, TypeVar

T = TypeVar("T")


class Indexable(Iterable, Sized, Generic[T]):
    """Indexable type for mypy."""

    def __getitem__(self, idx: int) -> T:
        """Get item at index idx."""
        ...  # pragma: no cover



from matplotlib import pyplot as plt
import numpy as np
from jacobi import jacobi
from numdifftools import Derivative


# function of one variable with auxiliary argument; returns a vector
def f(x):
    return np.sin(x) / x


def fp(x):
    return np.cos(x) / x - np.sin(x) / x**2


x = np.linspace(-10, 10, 200)
fpx = fp(x)
fpx1, fpxe1 = jacobi(f, x, diagonal=True)
fpx2 = Derivative(lambda p: f(x + p))(0)

plt.figure(constrained_layout=True)
plt.plot(x, np.abs(fpx1 - fpx), ls="-", label="Jacobi")
plt.plot(x, np.abs(fpx2 - fpx), ls="--", label="numdifftools")
plt.fill_between(x, 0, fpxe1, alpha=0.5, facecolor="C0", label="error estimate")
plt.title("$f(x) = sin(x)/x,\\quad f'(x) = cos(x) / x - sin(x) / x^2$")
plt.legend(ncol=3, loc="upper center", frameon=False)
plt.semilogy()
plt.ylim(1e-17, 1e-13)
plt.ylabel("$|f'_\\mathrm{num}(x) - f'(x)|$")
plt.axhline(np.finfo(float).resolution, color="k", ls="--")

plt.savefig("doc/_static/precision.svg")



# -*- coding: utf-8 -*-
#
# python_example documentation build configuration file, created by
# sphinx-quickstart on Fri Feb 26 00:29:33 2016.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

from pathlib import Path

with open("index.rst.in") as f:
    stub = f.read()

with open(Path("..") / "README.rst") as f:
    readme = f.read().replace(
        "https://hdembinski.github.io/jacobi/_images/", "_static/"
    )

with open("index.rst", "w") as f:
    f.write(stub + "\n" + readme)

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.doctest",
]

autosummary_generate = True

# General information about the project.
project = "Jacobi"
copyright = "2021, Hans Dembinski"
author = "Hans Dembinski"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
with open(Path() / ".." / "src" / "jacobi" / "_version.py") as f:
    d = {}
    exec(f.read(), d)

# The full version, including alpha/beta/rc tags.
release = d["version"]

# The short X.Y version.
version = ".".join(release.split(".")[:2])

language = "en"

exclude_patterns = ["_build"]

pygments_style = "sphinx"

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"

html_static_path = ["_static"]



from jacobi import jacobi
import numpy as np
from numdifftools import Derivative
from timeit import timeit
from matplotlib import pyplot as plt

fn = [
    "x ** 2",
    "exp(x)",
    "ones_like(x)",
    "mean(x ** 2)",
]

fig, ax = plt.subplots(
    2, 2, figsize=(10, 7), sharex=True, sharey=True, constrained_layout=True
)
for i, fi in enumerate(fn):
    t = {}
    t["jacobi"] = []
    t["numdifftools"] = []
    nmax = 5
    n = (10 ** np.linspace(0, nmax, nmax + 1)).astype(int)
    f = eval("lambda x: " + fi, np.__dict__)
    for ni in n:
        print(i, ni)
        x = np.linspace(0.1, 10, ni)
        number = 500 // ni + 10
        r = timeit(lambda: jacobi(f, x, diagonal=True), number=number) / number
        t["jacobi"].append(r)
        number = 500 // ni + 1
        r = timeit(lambda: Derivative(lambda p: f(p + x))(0), number=number) / number
        t["numdifftools"].append(r)
    plt.sca(ax.flat[i])
    for k, v in t.items():
        ls = "-" if "jacobi" in k else "--"
        plt.plot(n, v, ls=ls, label=k)
    r = np.divide(t["numdifftools"], t["jacobi"])
    plt.plot(n, r, ls=":", color="k", lw=3, label="time ratio")
    for ni, ri in zip(n, r):
        plt.text(
            ni, ri * 0.5, f"{ri:.0f}" if i != 3 else f"{ri:.1f}", va="top", ha="center"
        )
    plt.legend(loc="upper left", frameon=False)
    plt.title(fi)
plt.loglog()
fig.supxlabel("n")
fig.supylabel("t/sec")

plt.savefig("doc/_static/speed.svg")



from jacobi import jacobi
import pytest
import numpy as np

fn = ["0", "x ** 2", "x ** -1", "exp(x)", "sin(x)", "ones_like(x)", "mean(x ** 2)"]


@pytest.mark.parametrize("f", fn)
@pytest.mark.benchmark(group="1")
def test_jacobi_1(benchmark, f):
    f = eval(f"lambda x:{f}", np.__dict__)
    benchmark(jacobi, f, 0)


@pytest.mark.parametrize("f", fn)
@pytest.mark.benchmark(group="1000")
def test_jacobi_1000(benchmark, f):
    f = eval(f"lambda x:{f}", np.__dict__)
    x = np.linspace(0.1, 10, 1000)
    benchmark(jacobi, lambda p: f(x + p), 0)



"""Fast numerical derivatives for analytic functions with arbitrary round-off error."""

from ._propagate import propagate
from ._jacobi import jacobi
from ._version import version as __version__  # noqa

__all__ = ["jacobi", "propagate"]



from numpy.testing import assert_allclose, assert_equal
import numpy as np
import pytest
from jacobi import jacobi


def test_squeeze():
    # also tests promotion of integer argument to float
    y, ye = jacobi(np.exp, 0)
    assert_allclose(y, 1)
    assert_allclose(ye, 0, atol=1e-10)


def f1(x):
    return x**2 + 3


def fd1(x):
    return 2 * x


def f5(x):
    return np.mean(x**2)


def fd5(x):
    return 2 * x / len(x)


def f6(x):
    return np.outer(x, x)


def fd6(r):
    assert len(r) == 3
    x, y, z = r
    r = np.empty((3, 3, 3))
    r[..., 0] = [
        [2 * x, y, z],
        [y, 0, 0],
        [z, 0, 0],
    ]
    r[..., 1] = [
        [0, x, 0],
        [x, 2 * y, z],
        [0, z, 0],
    ]
    r[..., 2] = [
        [0, 0, x],
        [0, 0, y],
        [x, y, 2 * z],
    ]
    return r


def f7(x):
    return np.ones(3) * x**2


@pytest.mark.parametrize(
    "fn",
    [
        (lambda x: 1.0, lambda x: 0.0),
        (lambda x: np.exp(x), lambda x: np.diagflat(np.exp(x))),
        (lambda x: x**2, lambda x: np.diagflat(2 * x)),
        (lambda x: np.ones_like(x), lambda x: np.diagflat(np.zeros_like(x))),
        (lambda x: x**-1, lambda x: np.diagflat(-(x**-2))),
        (lambda x: (x + 1) ** 0.5, lambda x: np.diagflat(0.5 * (x + 1) ** -0.5)),
        (f5, fd5),
        (f6, fd6),
    ],
)
def test_jacobi(fn):
    x = np.array([1, 2, 3], dtype=float)
    f, fd = fn
    y, ye = jacobi(f, x)
    assert_allclose(y, fd(x))
    assert_allclose(ye, np.zeros_like(y), atol=1e-10)


def test_jacobi_0d():
    x = 2
    y, ye = jacobi(f7, x)
    assert np.ndim(y) == 1
    assert_allclose(y, f7(x))
    assert_allclose(ye, np.zeros_like(y), atol=1e-10)


def test_abs_at_zero():
    fp, fpe = jacobi(np.abs, 0)
    assert_equal(fp, 0)
    assert_equal(fpe, 0)


def test_method_auto():
    d = {}
    fp, fpe = jacobi(lambda x: x, 0, diagnostic=d)
    assert_equal(d["method"], [0])
    assert_allclose(fp, 1)
    assert_allclose(fpe, 0, atol=1e-10)

    fp, fpe = jacobi(lambda x: x if x >= 0 else np.nan, 0, diagnostic=d)
    assert_equal(d["method"], [1])
    assert_allclose(fp, 1)
    assert_allclose(fpe, 0, atol=1e-10)

    fp, fpe = jacobi(lambda x: x if x <= 0 else np.nan, 0, diagnostic=d)
    assert_equal(d["method"], [-1])
    assert_allclose(fp, 1)
    assert_allclose(fpe, 0, atol=1e-10)


def test_mask():
    x = np.array([1, 2, 3, 4])
    mask = np.array([True, False, False, True])
    jac = jacobi(f1, x, mask=mask)[0]

    assert jac.shape == (4, 2)
    jac_ref = np.diag(fd1(x))
    jac_ref = jac_ref[:, mask]

    assert_allclose(jac, jac_ref)


@pytest.mark.parametrize("method", (-1, 0, 1))
def test_method(method):
    d = {}
    fp, fpe = jacobi(lambda x: x, 0, method=method, diagnostic=d)
    assert_equal(d["method"], [method])
    assert_allclose(fp, 1)
    assert_allclose(fpe, 0, atol=1e-10)

    fp, fpe = jacobi(lambda x: x if x >= 0 else np.nan, 0, method=method, diagnostic=d)
    assert_equal(d["method"], [method])
    if method == 1:
        assert_allclose(fp, 1)
        assert_allclose(fpe, 0, atol=1e-10)
    else:
        assert_equal(fp, np.nan)
        assert_equal(fpe, np.inf)

    fp, fpe = jacobi(lambda x: x if x <= 0 else np.nan, 0, method=method, diagnostic=d)
    assert_equal(d["method"], [method])
    if method == -1:
        assert_allclose(fp, 1)
        assert_allclose(fpe, 0, atol=1e-10)
    else:
        assert_equal(fp, np.nan)
        assert_equal(fpe, np.inf)


@pytest.mark.parametrize("x", [2, [1, 2]])
def test_diagonal(x):
    jac = jacobi(f1, x, diagonal=True)[0]

    assert jac.ndim == np.ndim(x)
    assert_allclose(jac, fd1(np.atleast_1d(x)))


@pytest.mark.parametrize("maxgrad", (0, 2, 5))
def test_maxgrad(maxgrad):
    x = [1, 2, 3]

    jac = jacobi(f1, x, maxgrad=maxgrad)[0]

    jac_ref = np.diag(fd1(np.atleast_1d(x)))
    assert_allclose(jac, jac_ref)


def test_rtol():
    x = [1, 2, 3]

    jac, jace = jacobi(f1, x)

    jac2, jac2e = jacobi(f1, x, rtol=0.1)

    assert np.all(jac2e >= jace)
    assert_allclose(jac, jac2, rtol=0.02)


@pytest.mark.parametrize("maxiter", (0, -1))
def test_bad_maxiter(maxiter):
    with pytest.raises(ValueError):
        jacobi(f1, 1, maxiter=maxiter)


@pytest.mark.parametrize("maxgrad", (-1, -0.1))
def test_bad_maxgrad(maxgrad):
    with pytest.raises(ValueError):
        jacobi(f1, 1, maxgrad=maxgrad)


@pytest.mark.parametrize("step", [0, -1, 1, 0.5])
def test_bad_step_0(step):
    with pytest.raises(ValueError, match=r"step\[0\]"):
        jacobi(f1, 1, step=(step, 0.5))


@pytest.mark.parametrize("step", [0, -1, 1])
def test_bad_step_1(step):
    with pytest.raises(ValueError, match=r"step\[1\]"):
        jacobi(f1, 1, step=(0.25, step))


@pytest.mark.parametrize("method", (-2, -3, 2, 3))
def test_bad_method(method):
    with pytest.raises(ValueError, match="method"):
        jacobi(f1, 1, method=method)



from matplotlib import pyplot as plt
import numpy as np
from jacobi import jacobi


# function of one variable with auxiliary argument; returns a vector
def f(x):
    return np.sin(x) / x


x = np.linspace(-10, 10, 200)
fx = f(x)

# f(x) is a simple vectorized function, jacobian is diagonal
fdx, fdxe = jacobi(f, x, diagonal=True)
# fdxe is uncertainty estimate for derivative

plt.plot(x, fx, color="k", label="$f(x) = sin(x) / x$")
plt.plot(x, fdx, label="$f'(x)$ computed with jacobi")
scale = 14
plt.fill_between(
    x,
    fdx - fdxe * 10**scale,
    fdx + fdxe * 10**scale,
    label=f"$f'(x)$ error estimate$\\times \\, 10^{{{scale}}}$",
    facecolor="C0",
    alpha=0.5,
)
plt.legend()

plt.savefig("doc/_static/example.svg")
